<?php
// Класс бота

namespace BotKit\Common;

use BotKit\Events\Event;
use BotKit\Models\UserModel;
use BotKit\Events\PlainMessageEvent;
use BotKit\Events\CallbackEvent;
use BotKit\Enums\CallbackType;
use BotKit\Enums\State;

class Bot {

    // Событие которое обрабатывает бот
    private Event $event;

    // Загружен ли драйвер
    private bool $driver_loaded = false;

    // Загружает драйвер
    public function loadDriver($driver) {
        if ($this->driver_loaded == true) {
            // Драйвер уже выбран
            return;
        }

        if ($driver->forThis()) {
            $this->driver_loaded = true;
            $this->driver = $driver;
        }
    }

    // Убеждается в том, что драйвер для бота загружен
    public function ensureDriversLoaded() {
        // Выбросить исключение если ни один драйвер не согласился обработать
        // запрос
        if ($this->driver_loaded == false) {
            throw new \Exception("Bot has no loaded drivers");
        }
        $this->driver->onSelected();
        $this->event = $this->driver->getEvent();
    }

    // Общая процедура обработки запроса
    // $callback - то что будет выполняться
    // $params - доп. параметры для $callback
    private function processRequest(callable $callback, array $params) : void {
        $this->driver->onProcessStart();
        
        $user = $this->event->getUser();
        $callback_params = [
            'e' => $this->event,
            'u' => $user,
            'driver' => $this->driver
        ];
        $final_params = array_merge($callback_params, $params);
        call_user_func_array($callback, $final_params);

        // Сохранение пользователя...
        $em = Database::getEM();
        $em->flush();
        $this->driver->onUserSave($user);

        // Завершение работы драйвера...
        $this->driver->onProcessEnd();

        exit();
    }

    // Подключает обработчик события
    public function on(
        string $event_classname,
        callable $check,
        callable $callback
    ) : void
    {

        if (!is_a($this->event, $event_classname, true)) {
            // Если подключается обработчик события класса $event_classname,
            // а в этом запросе обрабатывается событие другого класса, то
            // и привязывать обработчик нет необходимости
            return;
        }

        // Проверка условия события
        $check_params = [
            'e' => $this->event,
            'u' => $this->event->getUser(),
            'driver' => $this->driver
        ];

        $result = call_user_func_array($check, $check_params);
        if ($result == false) {
            // Обрабатываемое событие - не для этого обработчика
            return;
        }
        $this->processRequest($callback, []);
    }

    // Подключает обработчик события без условия
    public function onEvent(string $event_classname, callable $callback) {
        if (!is_a($this->event, $event_classname, true)) {
            return;
        }
        $this->processRequest($callback, []);
    }

    // Подключает обработчик команды
    // Команда должна быть только текстовым сообщением
    public function onCommand(string $template, callable $callback) {
        if (!is_a($this->event, PlainMessageEvent::class, true)) {
            // Не текстовое сообщение
            return;
        }

        // Определяем что будет параметрами
		$pattern = '/^'.preg_replace(
			['/\//','/{(\w+)}/'],
			['\\\/', '(?<$1>.*)'],
			$template
		).'$/';


        if (!preg_match($pattern, $this->event->getText(), $named_groups)) {
            // Обрабатываемое событие - не для этого обработчика
            return;
		};

        // Оставляем только строковые ключи, т.к. в processRequest
        // $named_groups будут соединены со стандартными параметрами, которые
        // передаются только по названиям, не по позициям. Если в $named_groups
        // попадётся числовой ключ, позиционный аргумент передастся после
        // ключевого, что приведёт к ошибке
        $named_groups_filter = array_filter(
            $named_groups,
            function ($k) {
                return is_string($k);
            },
            ARRAY_FILTER_USE_KEY
        );

        // Вызов обработки с пойманными параметрами
        $this->processRequest($callback, $named_groups_filter);
    }

    // Подключает обработчик обратного вызова
    public function onCallback(CallbackType $callbackType, callable $responseCallback) {
        if (!is_a($this->event, CallbackEvent::class)) {
            // Не событие обратного вызова
            return;
        }

        if ($this->event->getCallbackType() != $callbackType) {
            // Этот тип обратного вызова не подходит
            return;
        }

        // Вызов обработки
        $this->processRequest($responseCallback, []);
    }

    // Подключает обработчик текстового для состояния пользователя
    public function onPlainMessage(State $required_state, callable $callback) {
        if (!$this->event->getUser()->inState($required_state)) {
            // Требуемое состояние у пользователя не обнаружено
            return;
        }
        $this->processRequest($callback, []);
    }

    // Все условия не прошли, вызываем план Б
    public function fallback(callable $callback) {
        $this->processRequest($callback, []);
    }
}