<?php
namespace BotKit\Drivers;

// Драйвер telegram.org

use BotKit\Events\Event;
use BotKit\Events\PlainMessageEvent;
use BotKit\Events\EmptyEvent;
use BotKit\Events\MemberJoinedEvent;
use BotKit\Events\MemberLeftEvent;
use BotKit\Common\User;
use BotKit\Common\Chat;
use BotKit\Common\Message;
use BotKit\Common\Database;
use BotKit\EventAttachments\DocumentAttachment;
use BotKit\EventAttachments\ImageAttachment;
use BotKit\Enums\Platform;
use BotKit\Enums\State;
use BotKit\Models\UserModel;

class TgBotDriver implements Driver {
    // Токен бота
    private string $bot_token;

    // Базовый адрес API
    private string $api_base;

    // Платформа бота
    private Platform $platform;

    public function __construct($bot_token) {
        $this->bot_token = $bot_token;
        $this->api_base = "https://api.telegram.org/bot".$this->bot_token;
        $this->platform = Platform::TelegramOrg;
    }

    public function forThis() : bool {
        // Событие от Телеграм содержит update_id
        $data = json_decode(file_get_contents("php://input"));
        if ($data == null) {
            // Нет входных данных - нет ответа
            return false;
        }
        return property_exists($data, 'update_id');
    }

    public function getEvent() : Event {
        $data = json_decode(file_get_contents("php://input"));
        $db = Database::getConnection();

        if (property_exists($data, 'message')) {
            $chat = new Chat($data->message->chat->id);
            
            if (property_exists($data->message, 'text')) {
                // Обычное текстовое сообщение

                // Получение объекта из БД
                $obj = UserModel::where([
                    ['platform_id', '=', $data->message->from->id],
                    ['platform', '=', $this->platform]
                ]);

                if ($obj === false) {
                    // Пользователя нет в БД
                    $obj = UserModel::create(
                        $data->message->from->id,
                        $this->platform
                    );
                }

                $user = new User(
                    $data->message->from->id,
                    State::Any,
                    $data->message->from->username,
                    $obj
                );
                $text = $data->message->text;

                $attachments = [];
                if (property_exists($data->message, 'document')) {
                    $attachments[] = new DocumentAttachment(
                        $data->message->document->file_id,
                        $data->message->document->file_name
                    );
                }
                if (property_exists($data->message, 'photo')) {
                    $attachments[] = new ImageAttachment(
                        $data->message->photo[0]->file_id
                    );
                }
                
                return new PlainMessageEvent(
                    $data->message->message_id,
                    $user,
                    $chat,
                    $text,
                    $attachments
                );
            }

            if (property_exists($data->message, 'new_chat_member')) {
                // Получение объекта из БД
                $obj = UserModel::where([
                    ['platform_id', '=', $data->message->new_chat_member->id],
                    ['platform', '=', $this->platform]
                ]);
                
                if ($obj === false) {
                    // Пользователя нет в БД
                    $obj = UserModel::create(
                        $data->message->from->id,
                        $this->platform
                    );
                }
                
                $user = new User(
                    $data->message->new_chat_member->id,
                    State::Any,
                    $data->message->new_chat_member->username,
                    $obj
                );
                return new MemberJoinedEvent($user, $chat);
            }
    
            if (property_exists($data->message, 'left_chat_member')) {
                // Получение объекта из БД
                $obj = UserModel::where([
                    ['platform_id', '=', $data->message->left_chat_member->id],
                    ['platform', '=', $this->platform]
                ]);

                if ($obj === false) {
                    // Пользователя нет в БД
                    $obj = UserModel::create(
                        $data->message->from->id,
                        $this->platform
                    );
                }
                
                $user = new User(
                    $data->message->left_chat_member->id,
                    State::Any,
                    $data->message->left_chat_member->username,
                    $obj
                );
                return new MemberLeftEvent($user, $chat);
            }
        }

        //~ return new EmptyEvent();
        throw new \Exception("Unknown event");
    }

    public function reply(PlainMessageEvent $e, Message $msg, bool $empathise = true) {
        if ($empathise) {
            $reply_to_id = $e->getMessageID();
        } else {
            $reply_to_id = -1;
        }
        $this->sendToChatInternal($e->getChat()->getID(), $msg, $reply_to_id);
    }

    public function sendMessage(User $u, Message $msg) {
        $this->sendToChatInternal($u->getPlatformID(), $msg, -1);
    }

    public function sendToChat(Chat $chat, Message $msg) {
        $this->sendToChatInternal($chat->getPlatformID(), $msg, -1);
    }

    // Отправляет сообщение в чат
    // $chat_id - ID чата в telegram
    // $msg - отправляемое сообщение
    // $reply_to_id - ID сообщения, на которого нужно ответить. Если -1, то
    //   отвечать не нужно
    private function sendToChatInternal($chat_id, Message $msg, int $reply_to_id) : int {
        $url = $this->api_base."/sendMessage?";
        
        $params = array(
            "chat_id" => $chat_id,
            "text" => $msg->getText()
        );

        if ($reply_to_id != -1) {
            $params["reply_parameters"] = json_encode([
                'message_id'=>$reply_to_id
            ]);
        }
        
        $response = file_get_contents($url.http_build_query($params));
        echo $response;
        // TODO: считывать ответ от telegram и возвращать ID сообщения
        return 0;
    }

    public function getUserNick(User $u) : string {
        // Ник уже был получен
        return $u->getNick();
    }
}